/*
 * Copyright (c) 2019-2024 VMware Carbon Black, all rights reserved
 */

// Package config -- contains the configuration system for the Last Contact Updater
//
//	Version: 1.0.
//	Contact: team-enki@vmware.com
package config

import (
	"fmt"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/sirupsen/logrus"
	"github.com/spf13/viper"
)

var (
	config *Configurations = &Configurations{}
	once   sync.Once
)

// Configurations -- last contact time application settings
//
// This structure is filled in from the following sources, in this order (lower item takes precedence):
//   - programmatic defaults using "viper.SetDefault(<key>, <value>)"
//   - config file (defined by pscweb.NewApp as "config.yaml", in the current directory or in "/config)"
//   - defined as an environment variable or programmatically using "os.Setenv(<key>, <value>)"
//   - overridden programmatically after configuration is filled in (by use of "viper.ReadInConfig()) by using "viper.Set(<key>, <value>)"
//
// NOTE: "current directory" currently assumes top of the repository!
//
// "viper.Set()" allows you to clear any configurations by using a value of ""
type Configurations struct { //nolint:maligned  // We don't care about alignment
	// KclConfiguration for the consumer container.
	Name                     string
	RegionName               string `mapstructure:"region"`
	StreamName               string `mapstructure:"lct-stream-name"`
	ShardsPerContainer       int    `mapstructure:"shards-per-container"`
	MaxRecords               int    `mapstructure:"max-records"`
	ShardsSyncIntervalMillis int    `mapstructure:"shard-sync-interval-millis"`
	FailoverTimeMillis       int    `mapstructure:"fail-over-time-millis"`
	LeaseRefreshPeriodMillis int    `mapstructure:"lease-refresh-period-millis"`
	ShardInitialPosition     int    `mapstructure:"shard-initial-position"`
	CheckpointTableName      string `mapstructure:"checkpoint-table-name"`
	AwsEndpoint              string `mapstructure:"aws-endpoint"` // Used only for testing with localstack
	EnvCategory              string `mapstructure:"env-category"`
	BatchSize                int    `mapstructure:"batch-size"`
	CbdEnvName               string `mapstructure:"cbdenvname"`

	// ESConfiguration elastic search client config.
	EsMasterNodes                   string `mapstructure:"es-master-nodes"`
	EsBatchSize                     string `mapstructure:"es-batch-size"`
	DeviceSuppressionTimeSecs       string `mapstructure:"device-suppression-time-secs"`
	BulkProcessorWorkerThreads      int    `mapstructure:"bulk-processor-worker-threads"`
	EsSniff                         bool   `mapstructure:"es-sniff"`
	EsScheme                        string `mapstructure:"es-scheme"`
	EsVersions                      string `mapstructure:"es-versions"`
	EsUser                          string `mapstructure:"es-user"`
	EsPassword                      string `mapstructure:"es-password"`
	EsBatchSizeEnvMap               map[string]int
	DeviceSuppressionDurationEnvMap map[string]time.Duration
	EsVersionEnvMap                 map[string]string

	// Asset Registry properties
	AssetRegistryPushEnabled     bool   `mapstructure:"asset-registry-push-enabled"`
	AssetRegistryStreamName      string `mapstructure:"asset-registry-stream-name"`
	AssetRegistryKinesisEndpoint string `mapstructure:"asset-registry-kinesis-endpoint"`

	// Asset Registry Kinesis Publisher Configuration
	PublisherID          string `mapstructure:"lct-publisherId"`
	ProducerBatchCount   int    `mapstructure:"ar-producer-batch-count"`
	AggregationBatchSize int    `mapstructure:"ar-producer-agg-batch-size"`
	FlushIntervalSec     int    `mapstructure:"ar-producer-flush-interval-sec"`
}

// IsLocalEnv reports whether this is a local env by looking at the config's EnvCategory
func (cfg *Configurations) IsLocalEnv() bool {
	return cfg.EnvCategory == "local"
}

func parseConfig(cfg *Configurations) (*Configurations, error) {
	if err := viper.Unmarshal(cfg); err != nil {
		return nil, fmt.Errorf("can not unmarshal config: %v", err)
	}

	err := processEsBatchSize(cfg)
	if err != nil {
		return nil, err
	}

	err = processDeviceSuppressionTimeSecs(cfg)
	if err != nil {
		return nil, err
	}

	err = processEsVersions(cfg)
	if err != nil {
		return nil, err
	}
	return cfg, nil
}

// GetConfig -- function to init the config
func GetConfig() *Configurations {
	once.Do(func() {
		_, err := parseConfig(config)
		if err != nil {
			logrus.Fatalf("error parsing config: %s", err)
		}
	})

	return config
}

// processEsBatchSize -- reads a comma-separated string of data in the form of <env>@<value>
//   - local@2000,dev01@3000
func processEsBatchSize(config *Configurations) error {
	config.EsBatchSizeEnvMap = make(map[string]int)
	splitBatchSize := strings.Split(config.EsBatchSize, ",")
	if strings.TrimSpace(config.EsBatchSize) == "" {
		return fmt.Errorf("'es-batch-size' was not defined")
	}
	for _, s := range splitBatchSize {
		splits := strings.Split(strings.TrimSpace(s), "@")
		if len(splits) < 2 {
			return fmt.Errorf("invalid 'es-batch-size' definition: '%s'", strings.TrimSpace(s))
		}
		bs, err := strconv.Atoi(splits[1])
		if err != nil {
			return fmt.Errorf("failed to parse 'es-batch-size': %v", err)
		}
		config.EsBatchSizeEnvMap[splits[0]] = bs
	}
	return nil
}

// processDeviceSuppressionTimeSecs -- reads a comma-separated string of data in the form of <env>@<sec>
//   - local@1000,dev01@3000
func processDeviceSuppressionTimeSecs(config *Configurations) error {
	config.DeviceSuppressionDurationEnvMap = make(map[string]time.Duration)
	splitSuppression := strings.Split(config.DeviceSuppressionTimeSecs, ",")
	if strings.TrimSpace(config.DeviceSuppressionTimeSecs) == "" {
		return fmt.Errorf("'device-suppression-time-secs' was not defined")
	}
	for _, s := range splitSuppression {
		splits := strings.Split(strings.TrimSpace(s), "@")
		if len(splits) < 2 {
			return fmt.Errorf("invalid 'device-suppression-time-secs' definition: '%s'", strings.TrimSpace(s))
		}
		suppSecs, err := strconv.Atoi(splits[1])
		if err != nil {
			return fmt.Errorf("failed to parse 'device-suppression-time-secs': %v", err)
		}
		config.DeviceSuppressionDurationEnvMap[splits[0]] = time.Duration(suppSecs) * time.Second
	}
	return nil
}

// processEsVersions -- reads a comma-separated list of data in the form of <env>@<es-version>
//   - local@es_6,dev01@es_56
func processEsVersions(config *Configurations) error {
	config.EsVersionEnvMap = make(map[string]string)
	splitEsVersions := strings.Split(config.EsVersions, ",")
	if strings.TrimSpace(config.EsVersions) == "" {
		return fmt.Errorf("'es-versions' was not defined")
	}
	for _, s := range splitEsVersions {
		splits := strings.Split(strings.TrimSpace(s), "@")
		if len(splits) < 2 {
			return fmt.Errorf("invalid 'es-versions' definition: '%s'", strings.TrimSpace(s))
		}
		ver := strings.TrimSpace(splits[1])
		if strings.TrimSpace(ver) == "" {
			return fmt.Errorf("'es-versions' definition has no value: '%s'", strings.TrimSpace(s))
		}
		config.EsVersionEnvMap[splits[0]] = ver
	}
	return nil
}
