/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package config

import (
	"os"
	"testing"

	"github.com/rs/zerolog"
	"github.com/spf13/viper"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/logger"
)

var log = zerolog.New(os.Stdout)

func setup() {
	viper.AllowEmptyEnv(true)
	viper.AutomaticEnv()
	viper.SetConfigName("config")
	viper.SetConfigType("yaml")
	viper.AddConfigPath("../..") // default to the one defined at the top of the repository

	// in the case of subsequent tests, make sure set entries are corrected
	viper.Set("es-batch-size", "local@1000")
	viper.Set("device-suppression-time-secs", "local@100")
	viper.Set("es-versions", "local@es_6")
}

func TestConfiguration(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	log.Info().Msgf("Config used: %s", viper.ConfigFileUsed())

	conf, err := parseConfig(&Configurations{})
	if err != nil {
		t.Errorf("Failed to get config as expected: %v", err)
	}
	log.Info().Msgf("Saw: %+v", conf)
	if conf.PublisherID != "psc:platform:ems:app:last_contact_time" {
		t.Errorf("Did not see expected 'PublisherID' of 'psc:platform:ems:app:last_contact_time'")
	}
}

func TestConfiguration_EsBatchSize_Missing(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_EsBatchSize_Missing")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("es-batch-size", "")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_EsBatchSize_Malformed(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_EsBatchSize_Malformed")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("es-batch-size", "local")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_EsBatchSize_Invalid(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_EsBatchSize_Invalid")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("es-batch-size", "local@BOGUS")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_DeviceSuppression_Missing(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_DeviceSuppression_Missing")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("device-suppression-time-secs", "")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_DeviceSuppression_Malformed(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_DeviceSuppression_Malformed")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("device-suppression-time-secs", "local")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_DeviceSuppression_Invalid(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_DeviceSuppression_Invalid")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("device-suppression-time-secs", "local@BOGUS")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_EsVersions_Missing(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_EsVersions_Missing")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("es-versions", "")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_EsVersions_Malformed(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_EsVersions_Malformed")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("es-versions", "local")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}

func TestConfiguration_EsVersions_Invalid(t *testing.T) {
	hf := logger.HeaderFunc("TestConfiguration_EsVersions_Invalid")
	defer hf()

	setup()
	err := viper.ReadInConfig()
	if err != nil {
		t.Errorf("Unable to read configuration: %v", err)
	}
	viper.Set("es-versions", "local@")

	_, err = parseConfig(&Configurations{})
	log.Info().Msgf("Saw Error: %v", err)
	if err == nil {
		t.Errorf("Failed to get expected error")
	}
}
