/*
 * Copyright (c) 2023-2024 VMware Carbon Black, all rights reserved
 */

package consumer

import (
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"
)

// KclMonitoringService sends kcl stats
type KclMonitoringService struct {
	log *zerolog.Logger
	*kclMetrics
}

type kclMetrics struct {
	processRecordsCount   stats.MetricUpdateFunc
	processRecordsBytes   stats.MetricUpdateFunc
	millisBehindLatest    stats.MetricUpdateFunc
	getRecordsLatency     stats.MetricUpdateFunc
	processRecordsLatency stats.MetricUpdateFunc
}

func newKclMetrics(promFactory *stats.PrometheusMetricsFactory) *kclMetrics {
	return &kclMetrics{
		processRecordsCount:   promFactory.NewCounterVec("processed_records_count", "kcl", []string{"shard"}),
		processRecordsBytes:   promFactory.NewGaugeVec("processed_records_bytes", "kcl", []string{"shard"}),
		millisBehindLatest:    promFactory.NewGaugeVec("millis_behind_latest", "kcl", []string{"shard"}),
		getRecordsLatency:     promFactory.NewGaugeVec("get_records_latency", "kcl", []string{"shard"}),
		processRecordsLatency: promFactory.NewGaugeVec("process_records_latency", "kcl", []string{"shard"}),
	}
}

func NewKclMonitoringService(promFactory *stats.PrometheusMetricsFactory, log *zerolog.Logger) *KclMonitoringService {
	return &KclMonitoringService{
		log:        log,
		kclMetrics: newKclMetrics(promFactory),
	}
}

// Init -- Initialize the service
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) Init(_, _, _ string) error { return nil }

// Start -- starts the service
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) Start() error { return nil }

// Shutdown -- Shutdown the service
//   - currently a no-op
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) Shutdown() {
}

// IncrRecordsProcessed -- update stats for processed records count
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) IncrRecordsProcessed(shard string, value int) {
	k.processRecordsCount(float64(value), shard)
}

// IncrBytesProcessed -- update stats for processed records bytes
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) IncrBytesProcessed(shard string, bytes int64) {
	k.processRecordsBytes(float64(bytes), shard)
}

// MillisBehindLatest -- update stats for consumer delay
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) MillisBehindLatest(shard string, millSeconds float64) {
	k.millisBehindLatest(float64(millSeconds), shard)
}

// LeaseGained -- update stats for lease gained
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) LeaseGained(shard string) {
	log.Warn().Msgf("Lease gained on shard with ID - %s", shard)
}

// LeaseLost -- update stats for lease lost
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) LeaseLost(shardID string) {
	log.Warn().Msgf("Lease lost on shard with ID - %s", shardID)
}

// LeaseRenewed  -- update stats for lease renewed
//   - currently only logs
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) LeaseRenewed(shard string) {
	log.Warn().Msgf("Lease renewed on shard with ID - %s", shard)
}

// RecordGetRecordsTime -- update stats for latency of get records
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) RecordGetRecordsTime(shard string, millis float64) {
	k.getRecordsLatency(millis, shard)
}

// RecordProcessRecordsTime -- update stats for latency of process records
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) RecordProcessRecordsTime(shard string, millis float64) {
	k.processRecordsLatency(millis, shard)
}
