/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

// Package consumer -- Receive device messages from kinesis stream
//
//	Version: 1.0.
//	Contact: team-enki@vmware.com
package consumer

import (
	cfg "github.com/vmware/vmware-go-kcl/clientlibrary/config"
	kc "github.com/vmware/vmware-go-kcl/clientlibrary/interfaces"
	wk "github.com/vmware/vmware-go-kcl/clientlibrary/worker"
	"github.com/vmware/vmware-go-kcl/logger"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/publisher"
)

// LctConsumer -- Used to fetch last contact time messages from kinesis stream and process them.
type LctConsumer struct {
	worker               *wk.Worker
	assetRegistryEnabled bool
	kinesisProducer      *publisher.KinesisPublisher
	log                  logger.Logger
}

// NewLctConsumer -- Create a new LctConsumer based on specified values
//   - primarily used for testing
func NewLctConsumer(log logger.Logger, worker *wk.Worker, assetRegistryEnabled bool, kinesisProducer *publisher.KinesisPublisher) LctConsumer {
	consumer := LctConsumer{
		worker:               worker,
		assetRegistryEnabled: assetRegistryEnabled,
		kinesisProducer:      kinesisProducer,
	}
	consumer.log = log
	return consumer
}

// NewLctConsumerFromKclConfig -- Create a new LctConsumer based on the kinesis configuration
func NewLctConsumerFromKclConfig(kclCfg *cfg.KinesisClientLibConfiguration, recordProcessorFactory kc.IRecordProcessorFactory, assetRegistryEnabled bool, kinesisProducer *publisher.KinesisPublisher) LctConsumer {
	worker := wk.NewWorker(recordProcessorFactory, kclCfg)
	return NewLctConsumer(
		kclCfg.Logger,
		worker,
		assetRegistryEnabled,
		kinesisProducer,
	)
}

// Start -- starts the worker for this consumer.
func (c *LctConsumer) Start() {
	err := c.worker.Start()
	if err != nil {
		c.log.Errorf("Failed to start consumer %v", err)
		panic(err)
	}

	if c.assetRegistryEnabled {
		c.kinesisProducer.StartProducer()
	}
}

// Stop -- stops the worker for this consumer.
func (c *LctConsumer) Stop() {
	c.worker.Shutdown()

	if c.assetRegistryEnabled {
		c.kinesisProducer.StopProducer()
	}
}
