/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package consumer

import (
	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"github.com/aws/aws-sdk-go/aws"
	"github.com/rs/zerolog"
	kc "github.com/vmware/vmware-go-kcl/clientlibrary/interfaces"
	"google.golang.org/protobuf/proto"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/publisher"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"
)

// Create a dump record processor for printing out all data from record.
type lctRecordProcessor struct {
	log                *zerolog.Logger
	config             *config.Configurations
	esBulkProcessorsV5 *publisher.ElasticSearchBulkProcessorsV5
	esBulkProcessorsV6 *publisher.ElasticSearchBulkProcessorsV6
	publishers         []*publisher.Publisher
	kinesisPublisher   *publisher.KinesisPublisher
	publisherMetrics   *publisher.PublisherMetrics
	esUpdateMetrics    *publisher.EsUpdateMetrics
}

// Initialize -- Initialize a lctRecordProcessor structure
func (p *lctRecordProcessor) Initialize(input *kc.InitializationInput) {
	p.log.Info().Msgf("Processing SharId: %v at checkpoint: %v", input.ShardId, aws.StringValue(input.ExtendedSequenceNumber.SequenceNumber))

	// The last contact time publisher is not thread-safe and thus created per instance of shard processor.
	lctPub := publisher.NewLCTPublisher(p.log, p.esBulkProcessorsV5, p.esBulkProcessorsV6, p.config, p.config.AssetRegistryPushEnabled, p.kinesisPublisher, p.esUpdateMetrics)
	p.publishers = append(p.publishers,
		publisher.NewPublisher(&publisher.Config{
			Log:             p.log,
			BatchSize:       p.config.BatchSize,
			SuppressionTime: p.config.DeviceSuppressionDurationEnvMap,
			ChannelSize:     10,
			Sinkf:           lctPub.Publish,
		}, p.publisherMetrics))
}

func (p *lctRecordProcessor) ProcessRecords(input *kc.ProcessRecordsInput) {
	// don't process empty record
	if len(input.Records) == 0 {
		return
	}

	for _, v := range input.Records {
		lastContactTime := &devicelastcontacttime.DeviceLastContactTime{}
		err := proto.Unmarshal(v.Data, lastContactTime)
		if err != nil {
			p.log.Error().Msgf("Unable to parse message received")
		}

		for _, r := range p.publishers {
			r.OnUpdate(lastContactTime)
		}
	}

	// checkpoint it after processing this batch
	lastRecordSequenceNumber := input.Records[len(input.Records)-1].SequenceNumber
	if input.MillisBehindLatest > 60_000 {
		p.log.Warn().Msgf("Checkpoint progress at: %v,  MillisBehindLatest = %v", lastRecordSequenceNumber, input.MillisBehindLatest)
	}
	err := input.Checkpointer.Checkpoint(lastRecordSequenceNumber)
	if err != nil {
		p.log.Error().Msgf("Checkpointing failed due to %v", err)
	}
}

func (p *lctRecordProcessor) Shutdown(input *kc.ShutdownInput) {
	p.log.Info().Msgf("Shutdown Reason: %v", aws.StringValue(kc.ShutdownReasonMessage(input.ShutdownReason)))

	// When the value of {@link ShutdownInput#getShutdownReason()} is
	// {@link com.amazonaws.services.kinesis.clientlibrary.lib.worker.ShutdownReason#TERMINATE} it is required that you
	// checkpoint. Failure to do so will result in an IllegalArgumentException, and the KCL no longer making progress.
	if input.ShutdownReason == kc.TERMINATE {
		err := input.Checkpointer.Checkpoint(nil)
		if err != nil {
			p.log.Error().Msgf("Encountered error while shutting down consumer %v", err)
			panic(err)
		}
	}
}

// LctRecordProcessorFactory is used to create RecordProcessor for last contact time messages
func LctRecordProcessorFactory(log *zerolog.Logger, conf *config.Configurations,
	bulkProcessorsV5 *publisher.ElasticSearchBulkProcessorsV5, bulkProcessorsV6 *publisher.ElasticSearchBulkProcessorsV6,
	producer *publisher.KinesisPublisher, metricsFactory *stats.PrometheusMetricsFactory,
) kc.IRecordProcessorFactory {
	return LctRecordProcessorFactoryWithMetrics(log, conf, bulkProcessorsV5, bulkProcessorsV6, producer, publisher.NewPublisherMetrics(metricsFactory), publisher.NewEsUpdateMetrics(metricsFactory))
}

// LctRecordProcessorFactoryWithMetrics is used to create RecordProcessor for last contact time messages
func LctRecordProcessorFactoryWithMetrics(log *zerolog.Logger, conf *config.Configurations,
	bulkProcessorsV5 *publisher.ElasticSearchBulkProcessorsV5, bulkProcessorsV6 *publisher.ElasticSearchBulkProcessorsV6,
	producer *publisher.KinesisPublisher, publisherMetrics *publisher.PublisherMetrics, esUpdateMetrics *publisher.EsUpdateMetrics,
) kc.IRecordProcessorFactory {
	return &lctRecordProcessorFactory{conf, log, bulkProcessorsV5, bulkProcessorsV6, producer, publisherMetrics, esUpdateMetrics}
}

// lct Record Processor factory
type lctRecordProcessorFactory struct {
	config             *config.Configurations
	log                *zerolog.Logger
	esBulkProcessorsV5 *publisher.ElasticSearchBulkProcessorsV5
	esBulkProcessorsV6 *publisher.ElasticSearchBulkProcessorsV6
	kinesisPublisher   *publisher.KinesisPublisher
	publisherMetrics   *publisher.PublisherMetrics
	esUpdateMetrics    *publisher.EsUpdateMetrics
}

func (d *lctRecordProcessorFactory) CreateProcessor() kc.IRecordProcessor {
	return &lctRecordProcessor{
		log:                d.log,
		config:             d.config,
		esBulkProcessorsV5: d.esBulkProcessorsV5,
		esBulkProcessorsV6: d.esBulkProcessorsV6,
		kinesisPublisher:   d.kinesisPublisher,
		publisherMetrics:   d.publisherMetrics,
		esUpdateMetrics:    d.esUpdateMetrics,
	}
}
