/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package consumer

import (
	"fmt"
	"net/http"
	"os"
	"os/signal"
	"strings"
	"syscall"

	"github.com/aws/aws-sdk-go/aws/credentials"
	"github.com/google/uuid"
	"github.com/labstack/echo/v4"
	"github.com/rs/zerolog"
	kclConfig "github.com/vmware/vmware-go-kcl/clientlibrary/config"
	"gitlab.bit9.local/psckit/pscweb"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/logger"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/publisher"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"
)

// StartApp starts the last contact updater app
func StartApp(conf *config.Configurations, log *zerolog.Logger, app *pscweb.WebApp) error {
	var kclConf *kclConfig.KinesisClientLibConfiguration

	if conf.IsLocalEnv() {
		// Running locally -- we need to specify some credentials to use
		creds := credentials.NewEnvCredentials()
		kclConf = kclConfig.NewKinesisClientLibConfigWithCredentials(conf.Name, conf.StreamName, conf.RegionName, uuid.New().String(), creds, creds).WithDynamoDBEndpoint(conf.AwsEndpoint).WithKinesisEndpoint(conf.AwsEndpoint)
	} else {
		kclConf = kclConfig.NewKinesisClientLibConfig(conf.Name, conf.StreamName, conf.RegionName, uuid.New().String())
	}
	log.Info().Msgf("Configuring %s for %s at %s", conf.Name, conf.StreamName, conf.RegionName)
	promMetricsFactory := stats.NewPrometheusMetrticsFactory(conf.CbdEnvName, log)

	kclConf = kclConf.WithTableName(conf.CheckpointTableName).
		WithFailoverTimeMillis(conf.FailoverTimeMillis).
		WithMaxRecords(conf.MaxRecords).
		WithLeaseRefreshPeriodMillis(conf.LeaseRefreshPeriodMillis).
		WithShardSyncIntervalMillis(conf.ShardsSyncIntervalMillis).
		WithMaxLeasesForWorker(conf.ShardsPerContainer).
		WithInitialPositionInStream(kclConfig.InitialPositionInStream(conf.ShardInitialPosition)).
		WithLogger(logger.NewKclLogger(log)).
		WithMonitoringService(NewKclMonitoringService(promMetricsFactory, log))
	kclConf.InitialLeaseTableReadCapacity = 10
	kclConf.InitialLeaseTableWriteCapacity = 10

	clients := InitElasticSearchClients(conf, log)
	esBulkProcessorsV5 := publisher.ElasticSearchBulkProcessorsV5{ClientsMap: clients, Log: log, Conf: conf}
	esBulkProcessorsV6 := publisher.ElasticSearchBulkProcessorsV6{ClientsMap: clients, Log: log, Conf: conf}

	// Connect to Asset Registry Kinesis stream if the Flag is enabled and check if stream exists
	var kinesisProducer *publisher.KinesisPublisher
	if conf.AssetRegistryPushEnabled {
		log.Info().Msgf("Initializing Kinesis Publisher for Asset Registry, stream = %v", conf.AssetRegistryStreamName)
		kinesisProducer = publisher.NewKinesisPublisher(log, conf, promMetricsFactory)
		if !kinesisProducer.StreamExists(&conf.AssetRegistryStreamName) {
			err := fmt.Errorf("asset registry stream enabled but '%v' doesn't exist", conf.AssetRegistryStreamName)
			return err
		}
	}

	app.MetricsRegistry.Start()
	recordProcessorFactory := LctRecordProcessorFactory(log, conf, &esBulkProcessorsV5, &esBulkProcessorsV6, kinesisProducer, promMetricsFactory)
	lctConsumer := NewLctConsumerFromKclConfig(kclConf, recordProcessorFactory, conf.AssetRegistryPushEnabled, kinesisProducer)
	lctConsumer.Start()
	log.Warn().Msgf("===== LCT consumer started =====")
	app.Port = 5004
	app.Echo.Any("/health_check", healthCheck())
	promMetricsFactory.RegisterGoMetrics()
	app.Echo.Any("/metrics", echo.WrapHandler(promMetricsFactory.HttpHandler()))
	app.Start()

	waitForSignal()
	lctConsumer.Stop()
	app.MetricsRegistry.Close()
	return nil
}

func healthCheck() func(c echo.Context) error {
	return func(c echo.Context) error {
		return c.String(http.StatusOK, "OK")
	}
}

// InitElasticSearchClients creates elastic search client connections for all the envs
func InitElasticSearchClients(conf *config.Configurations, log *zerolog.Logger) *publisher.ElasticSearchClientsMap {
	clientsMap := publisher.ElasticSearchClientsMap{}
	esNodesMap := make(map[string][]string)
	nodes := strings.Split(conf.EsMasterNodes, ",")
	for _, s := range nodes {
		envNode := strings.Split(s, "@")
		esNodesMap[envNode[0]] = append(esNodesMap[envNode[0]], envNode[1])
	}
	clientsMap.Initialize(conf, log, esNodesMap)
	return &clientsMap
}

func waitForSignal() {
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGQUIT, syscall.SIGTERM)
	<-sigs
}
