/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package integration

import (
	"fmt"
	"log"
	"net/http"
	"net/http/httptest"
	"net/url"
	"os"
	"sync/atomic"
	"testing"
	"time"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/service/kinesis"
	"github.com/rs/zerolog"
	kclcfg "github.com/vmware/vmware-go-kcl/clientlibrary/config"
	wk "github.com/vmware/vmware-go-kcl/clientlibrary/worker"
	kclog "github.com/vmware/vmware-go-kcl/logger" //nolint:goimports // IGNORE DSER-45076
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/consumer"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/publisher"
	"gitlab.bit9.local/psc/platform/orgkey-go/v3" //nolint:goimports // Test, IGNORE DSER-45076
	"google.golang.org/protobuf/proto"
)

const (
	lctStreamName           = "device-last-contact-time"
	assetRegistryStreamName = "asset-registry-stream"
)

var (
	testLogger                                       = zerolog.New(os.Stdout)
	metricsFactory   *stats.PrometheusMetricsFactory = stats.NewPrometheusMetrticsFactory("integration", &testLogger)
	kplMetrics       *publisher.KplMetrics           = publisher.NewKplMetrics(metricsFactory)
	esUpdateMetrics  *publisher.EsUpdateMetrics      = publisher.NewEsUpdateMetrics(metricsFactory)
	publisherMetrics *publisher.PublisherMetrics     = publisher.NewPublisherMetrics(metricsFactory)
)

// produceRecords -- Add records to the mock kinesis stream
//
//goland:noinspection GoMixedReceiverTypes
func (m mockKinesisClient) produceRecords(streamName *string, orgKey fmt.Stringer, timeMillis int64, numRecords int) {
	entries := preparePutRecords(orgKey, timeMillis, numRecords)
	m.addRecordsToKinesis(streamName, entries)
}

// preparePutRecords -- create a sequence of kinesis records for a set of devices
func preparePutRecords(orgKey fmt.Stringer, now int64, numRecords int) []*kinesis.PutRecordsRequestEntry {
	entries := make([]*kinesis.PutRecordsRequestEntry, numRecords)
	for i := 0; i < len(entries); i++ {
		data := devicelastcontacttime.DeviceLastContactTime{
			OrgKey:               orgKey.String(),
			DeviceId:             int64(i),
			LastContactTimestamp: now,
			DeregisterAfterTime:  now + 1000000,
		}
		//goland:noinspection GoVetCopyLock
		log.Printf("Adding record: OrgKey %s, DevId %d, LCT %d, DeReg %d", data.OrgKey, data.DeviceId, data.LastContactTimestamp, data.DeregisterAfterTime)
		marshal, _ := proto.Marshal(&data) // we expect no problems for known good data
		entries[i] = &kinesis.PutRecordsRequestEntry{
			Data:         marshal,
			PartitionKey: aws.String(orgKey.String()),
		}
	}
	return entries
}

////////////////////////////////////////////////////////////////////////////////
// Begin Tests
////////////////////////////////////////////////////////////////////////////////

// TestLctConsumer -- Use a mock kinesis stream to test our elasticsearch consumer.
func TestLctConsumer(t *testing.T) {
	// Wait for kinesis stream to initialize
	kc := newMockKinesisClient()
	if err := kc.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: aws.String(lctStreamName)}); err != nil {
		t.Errorf("Stream not found: %v", err)
	}

	// Create the kinesis consumer
	recordProcessorFactory := newMockRecordProcessorFactory()
	kclConfig := kclcfg.NewKinesisClientLibConfig("last-contact-updater", lctStreamName, "us-east-1", "worker")
	worker := wk.NewWorker(recordProcessorFactory, kclConfig).WithKinesis(kc).WithCheckpointer(mockCheckPointer{})
	kcLogger := kclog.GetDefaultLogger()
	lctConsumer := consumer.NewLctConsumer(kcLogger, worker, false, nil)

	// Create 10 devices, LCT to the current second, Deregister at +1,000,000 seconds
	now := time.Now().Unix() * 1000
	orgKey, _ := orgkey.Encode("local", 1) //nolint:errcheck // Fixed data, should never fail
	kc.produceRecords(aws.String(lctStreamName), orgKey, now, 10)

	// start the consumer, process for 2 seconds, then stop
	lctConsumer.Start()
	time.Sleep(2 * time.Second)
	lctConsumer.Stop()

	// validate
	if len(recordProcessorFactory.processor.processedRecords) != 10 {
		t.Fatalf("Failed to process all ten records")
	}
}

const rawdata = `{
			"took": 30,
			"errors": false,
			"items": [
			   {
				  "index": {
					 "_index": "a_devicestats",
					 "_type": "_doc",
					 "_id": "1",
					 "_version": 1,
					 "result": "created",
					 "_shards": {
						"total": 100,
						"successful": 100,
						"failed": 0
					 },
					 "status": 201,
					 "_seq_no" : 0,
					 "_primary_term": 1
				  }
			   },
			   {
				  "update": {
					 "_index": "a_devicestats",
					 "_type": "_doc",
					 "_id": "1",
					 "_version": 2,
					 "result": "updated",
					 "_shards": {
						 "total": 100,
						 "successful": 100,
						 "failed": 0
					 },
					 "status": 200,
					 "_seq_no" : 3,
					 "_primary_term" : 4
				  }
			   }
			]
		 }`

// Common function to get a moc kinesis client
func getKinesisClient(t *testing.T) *mockKinesisClient {
	kc := newMockKinesisClient()

	// Wait for kinesis streams to initialize
	if err := kc.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: aws.String(lctStreamName)}); err != nil {
		t.Errorf("Stream not found: %v", err)
	}
	if err := kc.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: aws.String(assetRegistryStreamName)}); err != nil {
		t.Errorf("Stream not found: %v", err)
	}

	return kc
}

func TestLctConsumerWithLctRecordProcessorFactoryEs6(t *testing.T) {
	// Get kinesis client
	kc := getKinesisClient(t)

	// create mock elasticsearch
	var numElasticUpdates int64
	mockElasticServer := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		atomic.AddInt64(&numElasticUpdates, 1)
		w.Header().Set("Content-Type", "application/json")
		w.WriteHeader(http.StatusOK)
		_, _ = w.Write([]byte(rawdata))
	}))
	mockServerURL, err := url.Parse(mockElasticServer.URL)
	if err != nil {
		t.Fatalf("Could not parse test elastic server url %v", err)
	}
	mockServerHostname := mockServerURL.Host
	defer mockElasticServer.Close()
	l := zerolog.New(os.Stdout)

	conf := &config.Configurations{
		EsBatchSizeEnvMap:        map[string]int{"local": 1},
		AssetRegistryStreamName:  assetRegistryStreamName,
		AssetRegistryPushEnabled: true,
		StreamName:               lctStreamName,
		EsVersions:               "local@es_6",
		EsMasterNodes:            fmt.Sprintf("local@%s", mockServerHostname),
		EsVersionEnvMap: map[string]string{
			"local": "es_6",
		},
		EsScheme:             "http",
		AggregationBatchSize: 100,
		ProducerBatchCount:   1,
		EsBatchSize:          "1",
	}

	clientMap := publisher.ElasticSearchClientsMap{}
	clientMap.Initialize(conf, &l, map[string][]string{"local": {
		mockServerHostname,
	}})

	esBulkProcessor := publisher.ElasticSearchBulkProcessorsV6{
		ClientsMap: &clientMap,
		Conf:       conf,
		Log:        &l,
	}

	kinesisPublisher := publisher.NewKinesisPublisherWithClient(kc, &l, conf, kplMetrics)
	recordProcessorFactory := consumer.LctRecordProcessorFactoryWithMetrics(&l, conf, nil, &esBulkProcessor, kinesisPublisher, publisherMetrics, esUpdateMetrics)

	kclConfig := kclcfg.NewKinesisClientLibConfig("last-contact-updater", "device-last-contact-time", "us-east-1", "worker")
	worker := wk.NewWorker(recordProcessorFactory, kclConfig).WithKinesis(kc).WithCheckpointer(mockCheckPointer{})
	kcLogger := kclog.GetDefaultLogger()
	lctConsumer := consumer.NewLctConsumer(kcLogger, worker, true, kinesisPublisher)

	// Create 10 devices, LCT to the current second, Deregister at +1,000,000 seconds
	now := time.Now().Unix() * 1000
	orgKey, _ := orgkey.Encode("local", 1)
	kc.produceRecords(aws.String(lctStreamName), orgKey, now, 10)

	// start the consumer, process for 10 seconds, then stop
	lctConsumer.Start()
	time.Sleep(10 * time.Second)
	lctConsumer.Stop()

	if atomic.LoadInt64(&numElasticUpdates) < 10 {
		t.Error("Failed to send updates to elasticsearch")
	}
	if len(kc.records[assetRegistryStreamName]) < 10 {
		t.Errorf("Didn't produce records to asset registry stream %v", kc.records)
	}
}

const rawData2 = `{
			"took": 30,
			"errors": false,
			"items": [
			   {
				  "index": {
					 "_index": "a_devicestats",
					 "_type": "_doc",
					 "_id": "1",
					 "_version": 1,
					 "result": "created",
					 "_shards": {
						"total": 100,
						"successful": 100,
						"failed": 0
					 },
					 "status": 201,
					 "_seq_no" : 0,
					 "_primary_term": 1
				  }
			   },
			   {
				  "update": {
					 "_index": "a_devicestats",
					 "_type": "_doc",
					 "_id": "1",
					 "_version": 2,
					 "result": "updated",
					 "_shards": {
						 "total": 100,
						 "successful": 100,
						 "failed": 0
					 },
					 "status": 200,
					 "_seq_no" : 3,
					 "_primary_term" : 4
				  }
			   }
			]
		 }`

func TestLctConsumerWithLctRecordProcessorFactoryEs5(t *testing.T) {
	// Get kinesis client
	kc := getKinesisClient(t)

	var numElasticUpdates int64
	mockElasticServer := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		atomic.AddInt64(&numElasticUpdates, 1)
		w.Header().Set("Content-Type", "application/json")
		w.WriteHeader(http.StatusOK)
		_, _ = w.Write([]byte(rawData2))
	}))
	mockServerURL, err := url.Parse(mockElasticServer.URL)
	if err != nil {
		t.Fatalf("Could not parse test elastic server url %v", err)
	}
	mockServerHostname := mockServerURL.Host
	defer mockElasticServer.Close()
	l := zerolog.New(os.Stdout)

	conf := &config.Configurations{
		EsBatchSizeEnvMap:        map[string]int{"local": 1},
		AssetRegistryStreamName:  assetRegistryStreamName,
		AssetRegistryPushEnabled: true,
		StreamName:               lctStreamName,
		EsVersions:               "local@es_56",
		EsMasterNodes:            fmt.Sprintf("local@%s", mockServerHostname),
		EsVersionEnvMap: map[string]string{
			"local": "es_56",
		},
		EsScheme:             "http",
		AggregationBatchSize: 100,
		ProducerBatchCount:   1,
		EsBatchSize:          "1",
	}

	clientMap := publisher.ElasticSearchClientsMap{}
	clientMap.Initialize(conf, &l, map[string][]string{"local": {
		mockServerHostname,
	}})

	esBulkProcessor := publisher.ElasticSearchBulkProcessorsV5{
		ClientsMap: &clientMap,
		Conf:       conf,
		Log:        &l,
	}

	kinesisPublisher := publisher.NewKinesisPublisherWithClient(kc, &l, conf, kplMetrics)
	recordProcessorFactory := consumer.LctRecordProcessorFactoryWithMetrics(&l, conf, &esBulkProcessor, nil, kinesisPublisher, publisherMetrics, esUpdateMetrics)

	kclConfig := kclcfg.NewKinesisClientLibConfig("last-contact-updater", "device-last-contact-time", "us-east-1", "worker")
	worker := wk.NewWorker(recordProcessorFactory, kclConfig).WithKinesis(kc).WithCheckpointer(mockCheckPointer{})
	kcLogger := kclog.GetDefaultLogger()
	lctConsumer := consumer.NewLctConsumer(kcLogger, worker, true, kinesisPublisher)

	// Create 10 devices, LCT to the current second, Deregister at +1,000,000 seconds
	now := time.Now().Unix() * 1000
	orgKey, _ := orgkey.Encode("local", 1)
	kc.produceRecords(aws.String(lctStreamName), orgKey, now, 10)

	// start the consumer, process for 10 seconds, then stop
	lctConsumer.Start()
	time.Sleep(10 * time.Second)
	lctConsumer.Stop()

	if atomic.LoadInt64(&numElasticUpdates) < 10 {
		t.Error("Failed to send updates to elasticsearch")
	}
	if len(kc.records[assetRegistryStreamName]) < 10 {
		t.Errorf("Didn't produce records to asset registry stream %v", kc.records)
	}
}
