//go:build ITs
// +build ITs

/*
 * Copyright (c) 2024 VMware Carbon Black, all rights reserved
 */

package integration

import (
	"context"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"os"
	"strconv"
	"testing"
	"time"

	"github.com/aws/aws-sdk-go/aws/credentials"
	"github.com/rs/zerolog"

	"github.com/spf13/viper"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/kinesis"
	"google.golang.org/protobuf/proto"
	"gopkg.in/olivere/elastic.v5"

	"gitlab.bit9.local/psc/platform/orgkey-go/v3"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/consumer"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/publisher"
)

// nolint
func TestLctConsumer56(t *testing.T) {
	hf := logger.TestId("TestLctConsumer56")
	defer hf()

	vars := setup2(t)
	defer cleanUp2(t, vars)

	// Sleeping for 20 secs to let the consumer app initialize.
	time.Sleep(time.Second * 10)

	// Create ES index and populate initial data

	err := createIndex2(vars.indexName, vars)
	if err != nil {
		panic(err)
	}
	orgKey, err := orgkey.Encode("local", 1)
	if err != nil {
		panic(err)
	}
	populateInitialDeviceData2(orgKey, vars, t)

	// Add 1 update for each device
	now := time.Now().Unix() * 1000
	produceRecords2(vars, orgKey, now)

	time.Sleep(time.Second * 2)
	esClient := vars.esClientV5
	res, err := esClient.Search(vars.indexName).Type("device").Do(context.Background())
	if err != nil {
		t.Error(err)
	}

	for _, hit := range res.Hits.Hits {
		verifyLastContactTime2(t, hit, now)
	}

	// Verify  suppression
	now2 := time.Now().Unix() * 1000
	produceRecords2(vars, orgKey, now2)
	time.Sleep(time.Second * 2)
	res, err = esClient.Search(vars.indexName).Type("device").Do(context.Background())
	if err != nil {
		t.Error(err)
	}
	// Should still be now as the update is suppressed for 5 secs.
	verifyLastContactTime2(t, res.Hits.Hits[0], now)

	// Verify  update after suppression time
	time.Sleep(time.Second * 2)
	now3 := time.Now().Unix() * 1000
	produceRecords2(vars, orgKey, now3)
	time.Sleep(time.Second * 2)
	res, err = esClient.Search(vars.indexName).Type("device").Do(context.Background())
	if err != nil {
		t.Error(err)
	}
	// Should still be now as the update is suppressed for 5 secs.
	verifyLastContactTime2(t, res.Hits.Hits[0], now3)
}

////////////////////////////////////////////////////////////////////////////////
// Support methods
////////////////////////////////////////////////////////////////////////////////

type TestVars struct {
	esClientV5 *elastic.Client
	conf       *config.Configurations
	streamName *string
	kc         *kinesis.Kinesis
	logger     zerolog.Logger
	indexName  string
}

func setup2(t *testing.T) *TestVars {
	logger := zerolog.New(os.Stderr).Level(zerolog.InfoLevel).With().Timestamp().Logger()
	viper.Set("dev", "true")
	conf := &config.Configurations{}
	setConfig2(conf)
	logger.Info().Msgf("Using configuration %+v", conf)
	esClientV5 := createEsClientV5(conf, logger)

	// Wait for kinesis stream in
	kc := createKinesisClient2(conf, t)
	streamName := aws.String("device-last-contact-time")
	if err2 := kc.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: streamName}); err2 != nil {
		t.Errorf("Stream not found: %v", err2)
	}

	return &TestVars{
		esClientV5: esClientV5,
		conf:       conf,
		streamName: streamName,
		kc:         kc,
		logger:     zerolog.Logger{},
		indexName:  "a_devicestats",
	}
}

// nolint
func createKinesisClient2(conf *config.Configurations, t *testing.T) *kinesis.Kinesis {
	newSession, err := session.NewSession(
		&aws.Config{
			Region:      aws.String(conf.RegionName),
			Endpoint:    aws.String(conf.AwsEndpoint),
			Credentials: credentials.NewStaticCredentials("test", "test", "test"),
		})
	if err != nil {
		t.Errorf("Could not connect to localstack: %v", err)
	}
	kc := kinesis.New(newSession)
	return kc
}

// nolint
func produceRecords2(vars *TestVars, orgKey *orgkey.OrgKey, timeMillis int64) {
	entries := preparePutRecords2(orgKey, timeMillis)
	addRecordsToKinesis2(entries, vars)
}

// nolint
func addRecordsToKinesis2(entries []*kinesis.PutRecordsRequestEntry, vars *TestVars) {
	_, err := vars.kc.PutRecords(&kinesis.PutRecordsInput{
		Records:    entries,
		StreamName: vars.streamName,
	})
	if err != nil {
		panic(err)
	}
}

// nolint
func setConfig2(conf *config.Configurations) {
	getAwsEndpointFromEnvOrDefault := func() string {
		if os.Getenv("AWS_ENDPOINT") != "" {
			return os.Getenv("AWS_ENDPOINT")
		}
		return "http://localhost:4566"
	}
	getEsMasterNodesFromEnvOrDefault := func() string {
		if os.Getenv("ES_MASTER_NODES") != "" {
			return os.Getenv("ES_MASTER_NODES")
		}
		return "local@localhost:9200"
	}

	conf.AwsEndpoint = getAwsEndpointFromEnvOrDefault()
	conf.EsMasterNodes = getEsMasterNodesFromEnvOrDefault()
	conf.EsVersionEnvMap = make(map[string]string)
	conf.EsVersionEnvMap[orgkey.Local] = "es_56"
	conf.EsSniff = false
	conf.EsScheme = "http"
	conf.RegionName = "us-east-1"
}

// nolint
func verifyLastContactTime2(t *testing.T, rec *elastic.SearchHit, now int64) {
	bytes, err := rec.Source.MarshalJSON()
	if err != nil {
		t.Errorf("Failed to unmarshall record %v", err)
	}
	stub := publisher.DeviceStatsStub{}
	fmt.Println(string(bytes))
	_ = json.Unmarshal(bytes, &stub)
	if stub.LastContact != now {
		t.Errorf("DeviceId %v lastContact is %v while expected is %v", stub, stub.LastContact, now)
	}
}

// nolint
func preparePutRecords2(orgKey *orgkey.OrgKey, now int64) []*kinesis.PutRecordsRequestEntry {
	entries := make([]*kinesis.PutRecordsRequestEntry, 10)
	for i := 0; i < len(entries); i++ {
		data0 := devicelastcontacttime.DeviceLastContactTime{
			OrgKey:               orgKey.String(),
			DeviceId:             int64(i),
			LastContactTimestamp: now,
			DeregisterAfterTime:  0,
		}
		// nolint
		marshal, _ := proto.Marshal(&data0)
		entries[i] = &kinesis.PutRecordsRequestEntry{
			Data:         marshal,
			PartitionKey: aws.String(orgKey.String()),
		}
	}
	return entries
}

// nolint
func populateInitialDeviceData2(orgKey *orgkey.OrgKey, vars *TestVars, t *testing.T) {
	orgIDStr := strconv.Itoa(orgKey.OrgID())
	for i := 0; i < 10; i++ {
		stub := publisher.DeviceStatsStub{
			OrganizationID: int64(orgKey.OrgID()),
			DeviceID:       int64(i),
		}
		_, err := vars.esClientV5.Index().Index(vars.indexName).
			Type("device").
			Id(fmt.Sprintf("%s-%s", orgIDStr, strconv.Itoa(int(stub.DeviceID)))).
			Routing(orgIDStr).
			BodyJson(&stub).
			Do(context.Background())
		if err != nil {
			t.Error(err)
		}
	}
}

// nolint
func createIndex2(indexName string, vars *TestVars) error {
	esClient := vars.esClientV5
	exists, err := esClient.IndexExists(indexName).Do(context.Background())
	if err != nil {
		return err
	}
	if exists == true {
		return nil
	}

	m, err := ioutil.ReadFile("../../docker/dev/es/devicestats_mapping_es56.json")
	mapping := string(m)
	if err != nil {
		return err
	}
	s, err := ioutil.ReadFile("../../docker/dev/es/devicestats_settings_es56.json")
	settings := string(s)
	if err != nil {
		return err
	}
	_, err = esClient.CreateIndex(indexName).BodyString(settings).Do(context.Background())
	if err != nil {
		return err
	}
	_, err = esClient.PutMapping().Index(indexName).Type("device").BodyString(mapping).Do(context.Background())
	if err != nil {
		return err
	}
	return nil
}

func createEsClientV5(conf *config.Configurations, logger zerolog.Logger) *elastic.Client {
	//nolint
	// FIXME: Figure out some way to remove these as the integration tests shouldn't have any src code dependencies
	esclientsMap := consumer.InitElasticSearchClients(conf, &logger)
	esclient := esclientsMap.GetEsClientV5("local")
	return esclient
}

func cleanUp2(t *testing.T, vars *TestVars) {
	// delete indexes
	_, err := vars.esClientV5.DeleteIndex(vars.indexName).Do(context.Background())
	if err != nil {
		t.Errorf("Failed to delete index %s, got error %v", vars.indexName, err)
	}
}
