//go:build ITs
// +build ITs

/*
 * Copyright (c) 2024 VMware Carbon Black, all rights reserved
 */

package integration

import (
	"context"
	"fmt"
	"log"
	"os"
	"testing"
	"time"

	"carbonblack.com/protocols/psc/ems/assetupdater"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/service/kinesis"
	"github.com/rs/zerolog"
	"github.com/spf13/viper"
	"google.golang.org/protobuf/proto"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
	"gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

//nolint:gocyclo
func TestLctConsumerKinesis(t *testing.T) {
	hf := logger.TestId("TestLctConsumerKinesis")
	defer hf()

	vars := setup3(t)
	defer cleanUp2(t, vars)

	// Sleeping for 20 secs to let the consumer app initialize.
	time.Sleep(time.Second * 10)

	// Create ES index and populate initial data
	err := createIndex2(vars.indexName, vars)
	if err != nil {
		panic(err)
	}
	orgKey, err := orgkey.Encode("local", 1)
	if err != nil {
		panic(err)
	}
	populateInitialDeviceData2(orgKey, vars, t)

	// Add 1 update for each device
	now := time.Now().Unix() * 1000
	produceRecords2(vars, orgKey, now)

	time.Sleep(time.Second * 2)
	esClient := vars.esClientV5
	res, err := esClient.Search(vars.indexName).Type("device").Do(context.Background())
	if err != nil {
		t.Error(err)
	}

	for _, hit := range res.Hits.Hits {
		verifyLastContactTime2(t, hit, now)
	}

	// Wait for Asset Registry kinesis stream init.
	kc := createKinesisClient2(vars.conf, t)

	assetRegistryStreamName := aws.String("asset-updater-input-stream")
	streams := describeStream(t, kc, assetRegistryStreamName)
	shardIterator := getShardIterator(kc, streams)
	var a *string

	// get data using infinity looping
	// we will attempt to consume data every 1 seconds, if no data, nothing will happen
	loopCount := 1
	for {
		fmt.Printf("Trying to read Kinesis records. Loop count = %d\n", loopCount)
		// get records use shard iterator for making request
		records, err := kc.GetRecords(&kinesis.GetRecordsInput{
			ShardIterator: shardIterator,
		})
		// if error, wait until 2 seconds and continue the looping process
		if err != nil {
			time.Sleep(2000 * time.Millisecond)
			continue
		}

		// process the data
		if len(records.Records) > 0 {
			for _, d := range records.Records {
				m := &assetupdater.AssetDataChangeEvent{}
				err1 := proto.Unmarshal(d.Data, m)
				if err1 != nil {
					t.Errorf("Error while marshaling Asset Registry Kinesis data - %v\n", err1)
					continue
				}
				log.Printf("GetRecords Data: %v\n", m)
			}
			fmt.Printf("Asset Registry Kinesis records count = %d\n", len(records.Records))
			break
		} else if records.NextShardIterator == a || shardIterator == records.NextShardIterator || err != nil {
			t.Errorf("GetRecords ERROR: %v\n", err)
			break
		}
		shardIterator = records.NextShardIterator
		time.Sleep(5000 * time.Millisecond)
		loopCount++
		// run the loop for 10 times max
		if loopCount == 11 {
			break
		}
	}

	// verify 10 records in Asset registry Stream
}

func getShardIterator(kc *kinesis.Kinesis, streams *kinesis.DescribeStreamOutput) *string {
	// retrieve iterator
	iteratorOutput, err := kc.GetShardIterator(&kinesis.GetShardIteratorInput{
		// Shard Id is provided when making put record(s) request.
		ShardId:           aws.String(*streams.StreamDescription.Shards[0].ShardId),
		ShardIteratorType: aws.String("TRIM_HORIZON"),
		// ShardIteratorType: aws.String("AT_SEQUENCE_NUMBER"),
		// ShardIteratorType: aws.String("LATEST"),
		StreamName: streams.StreamDescription.StreamName,
	})
	if err != nil {
		log.Panic(err)
	}

	shardIterator := iteratorOutput.ShardIterator
	return shardIterator
}

func describeStream(t *testing.T, kc *kinesis.Kinesis, streamName *string) *kinesis.DescribeStreamOutput {
	if err2 := kc.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: streamName}); err2 != nil {
		t.Errorf("Asset Registry Input Stream not found: %v", err2)
	}

	streams, err := kc.DescribeStream(&kinesis.DescribeStreamInput{StreamName: streamName})
	if err != nil {
		log.Panic(err)
	}
	return streams
}

func setup3(t *testing.T) *TestVars {
	logger := zerolog.New(os.Stderr).Level(zerolog.InfoLevel).With().Timestamp().Logger()
	viper.Set("dev", "true")
	conf := &config.Configurations{}
	setConfig3(conf)
	logger.Info().Msgf("Using configuration %+v", conf)
	esClientV5 := createEsClientV5(conf, logger)

	// Wait for kinesis stream init.
	lctConsumer := createKinesisClient2(conf, t)
	streamName := aws.String("device-last-contact-time")
	if err2 := lctConsumer.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: streamName}); err2 != nil {
		t.Errorf("Stream not found: %v", err2)
	}

	return &TestVars{
		esClientV5: esClientV5,
		conf:       conf,
		streamName: streamName,
		kc:         lctConsumer,
		logger:     zerolog.Logger{},
		indexName:  "a_devicestats",
	}
}

// nolint
func setConfig3(conf *config.Configurations) {
	getAwsEndpointFromEnvOrDefault := func() string {
		if os.Getenv("AWS_ENDPOINT") != "" {
			return os.Getenv("AWS_ENDPOINT")
		}
		return "http://localhost:4566"
	}
	getEsMasterNodesFromEnvOrDefault := func() string {
		if os.Getenv("ES_MASTER_NODES") != "" {
			return os.Getenv("ES_MASTER_NODES")
		}
		return "local@localhost:9200"
	}

	conf.AwsEndpoint = getAwsEndpointFromEnvOrDefault()
	conf.EsMasterNodes = getEsMasterNodesFromEnvOrDefault()
	conf.EsVersionEnvMap = make(map[string]string)
	conf.EsVersionEnvMap[orgkey.Local] = "es_56"
	conf.EsSniff = false
	conf.EsScheme = "http"
	conf.RegionName = "us-east-1"
	conf.AssetRegistryPushEnabled = true
	conf.AssetRegistryStreamName = "asset-updater-input-stream"
	conf.AggregationBatchSize = 1
	conf.ProducerBatchCount = 1
}
