/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package integration

import (
	"fmt"
	"log"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/request"
	"github.com/aws/aws-sdk-go/service/kinesis"
	"github.com/aws/aws-sdk-go/service/kinesis/kinesisiface"
	"github.com/vmware/vmware-go-kcl/clientlibrary/checkpoint"
	kciface "github.com/vmware/vmware-go-kcl/clientlibrary/interfaces"
	par "github.com/vmware/vmware-go-kcl/clientlibrary/partition"
)

const (
	shardID                = "shardId-00000000000"
	startingSequenceNumber = "start"
	endingSequenceNumber   = "stop"
)

// mockKinesisClient -- Structure use to feed the mock kinesis
type mockKinesisClient struct {
	kinesisiface.KinesisAPI
	records map[string][][]byte
}

// addRecordsToKinesis -- Add multiple records to the mock kinesis stream
//
//goland:noinspection GoMixedReceiverTypes
func (m mockKinesisClient) addRecordsToKinesis(streamName *string, entries []*kinesis.PutRecordsRequestEntry) {
	_, err := m.PutRecords(&kinesis.PutRecordsInput{
		Records:    entries,
		StreamName: streamName,
	})
	if err != nil {
		panic(err)
	}
}

//goland:noinspection GoMixedReceiverTypes
func (m *mockKinesisClient) WaitUntilStreamExists(input *kinesis.DescribeStreamInput) error {
	m.records[*input.StreamName] = make([][]byte, 0)
	return nil
}

//goland:noinspection GoMixedReceiverTypes
func (m *mockKinesisClient) GetShardIterator(input *kinesis.GetShardIteratorInput) (*kinesis.GetShardIteratorOutput, error) {
	streamName := input.StreamName
	return &kinesis.GetShardIteratorOutput{ShardIterator: streamName}, nil
}

//goland:noinspection GoMixedReceiverTypes
func (m *mockKinesisClient) GetRecords(input *kinesis.GetRecordsInput) (*kinesis.GetRecordsOutput, error) {
	ourOutput := m.records[*input.ShardIterator]
	outputLen := len(ourOutput)
	log.Printf("Get records returning %d", outputLen)
	outputRecords := make([]*kinesis.Record, outputLen)
	for i, recordBytes := range ourOutput {
		outputRecords[i] = &kinesis.Record{Data: recordBytes}
	}
	return &kinesis.GetRecordsOutput{Records: outputRecords, MillisBehindLatest: aws.Int64(0)}, nil
}

//goland:noinspection GoMixedReceiverTypes
func (m *mockKinesisClient) PutRecords(input *kinesis.PutRecordsInput) (output *kinesis.PutRecordsOutput, err error) {
	streamName := input.StreamName
	log.Printf("Producing records to stream : %s", *streamName)
	streamRecords := m.records[*streamName]
	lenInput := len(input.Records)
	newData := make([][]byte, lenInput)
	outputMessages := make([]*kinesis.PutRecordsResultEntry, lenInput)
	for index, record := range input.Records {
		newData[index] = record.Data
		outputMessages[index] = &kinesis.PutRecordsResultEntry{ShardId: aws.String(shardID), SequenceNumber: aws.String(fmt.Sprintf("%d", index))}
	}
	m.records[*streamName] = append(streamRecords, newData...)
	return &kinesis.PutRecordsOutput{Records: outputMessages, FailedRecordCount: aws.Int64(0)}, nil
}

//goland:noinspection GoMixedReceiverTypes
func (m *mockKinesisClient) PutRecordRequest(input *kinesis.PutRecordInput) (*request.Request, *kinesis.PutRecordOutput) {
	m.records[*input.StreamName] = append(m.records[*input.StreamName], input.Data)
	return &request.Request{}, nil
}

//goland:noinspection GoMixedReceiverTypes
func (m *mockKinesisClient) ListShards(*kinesis.ListShardsInput) (*kinesis.ListShardsOutput, error) {
	return &kinesis.ListShardsOutput{Shards: []*kinesis.Shard{{ShardId: aws.String(shardID), SequenceNumberRange: &kinesis.SequenceNumberRange{EndingSequenceNumber: aws.String(endingSequenceNumber), StartingSequenceNumber: aws.String(startingSequenceNumber)}}}}, nil
}

func newMockKinesisClient() *mockKinesisClient {
	return &mockKinesisClient{records: make(map[string][][]byte)}
}

func newMockRecordProcessorFactory() mockRecordProcessorFactory {
	return mockRecordProcessorFactory{processor: newMockRecordProcessor()}
}

type mockRecordProcessorFactory struct {
	kciface.IRecordProcessorFactory
	processor *mockRecordProcessor
}

func (m mockRecordProcessorFactory) CreateProcessor() kciface.IRecordProcessor {
	return m.processor
}

func newMockRecordProcessor() *mockRecordProcessor {
	return &mockRecordProcessor{processedRecords: make([]*kinesis.Record, 0)}
}

type mockRecordProcessor struct {
	kciface.IRecordProcessor
	processedRecords []*kinesis.Record
}

func (m *mockRecordProcessor) Initialize(_ *kciface.InitializationInput) {
}

func (m *mockRecordProcessor) ProcessRecords(processRecordsInput *kciface.ProcessRecordsInput) {
	log.Printf("PROCESSING RECORDS %d", len(processRecordsInput.Records))
	m.processedRecords = append(m.processedRecords, processRecordsInput.Records...)
}

func (m *mockRecordProcessor) Shutdown(_ *kciface.ShutdownInput) {
}

type mockCheckPointer struct {
	checkpoint.Checkpointer
}

func (m mockCheckPointer) CheckpointSequence(*par.ShardStatus) error {
	return nil
}

func (m mockCheckPointer) Init() error {
	return nil
}

func (m mockCheckPointer) RemoveLeaseOwner(string) error {
	return nil
}

func (m mockCheckPointer) FetchCheckpoint(*par.ShardStatus) error {
	return nil
}

func (m mockCheckPointer) GetLease(*par.ShardStatus, string) error {
	return nil
}

//
////////////////////////////////////////////////////////////////////////////////
