/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package logger

import (
	"fmt"
	"os"
	"testing"

	"github.com/rs/zerolog"
	"github.com/vmware/vmware-go-kcl/logger"
)

// Padding is defined as
func TestBannerPaddingUnder80(t *testing.T) {
	hf := HeaderFunc("TestBannerPaddingUnder80")
	defer hf()

	// ensure padding works (80, minus 2 spaces, minus length of tag, minus 5 equal signs
	source := "10 Chars.."
	check := padding(source)
	if len(check) != 80-17 {
		t.Errorf("Padding failed: %d chars\n%s", len(check), check)
	}
}

func TestBannerPaddingTagOver68(t *testing.T) {
	hf := HeaderFunc("TestBannerPaddingTagOver68")
	defer hf()

	// ensure we always have a pad of at least 5
	//goland:noinspection SpellCheckingInspection
	source := "70 Chars..xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
	check := padding(source)
	if len(check) != 5 {
		t.Errorf("Padding failed: %d chars\n%s", len(check), check)
	}
}

// NOTE: The following tests ensure our logging system is not breaking anything (and ups coverage)

func TestEsLogger(_ *testing.T) {
	hf := HeaderFunc("TestEsLogger")
	defer hf()

	log := zerolog.New(os.Stdout)
	esl := EsLogger{&log, zerolog.ErrorLevel}
	esl.Printf("Test Error")
	esl = EsLogger{&log, zerolog.WarnLevel}
	esl.Printf("Test Warn")
	esl = EsLogger{&log, zerolog.InfoLevel}
	esl.Printf("Test Info")
	esl = EsLogger{&log, zerolog.DebugLevel}
	esl.Printf("Test Debug")
}

func TestKclLogger(_ *testing.T) {
	// NOTE: Can't test Panicf or Fatalf since these kill (and fail) the test
	hf := HeaderFunc("TestKclLogger")
	defer hf()

	log := zerolog.New(os.Stdout)
	kcl := NewKclLogger(&log)

	kcl.Errorf("Test Errorf")
	kcl.Warnf("Test Warnf")
	kcl.Infof("Test Infof")
	kcl.Debugf("Test Debugf")

	kcl.WithFields(logger.Fields{}).Infof("Info with fields")
}

type logHookEvent struct {
	event *zerolog.Event
	level zerolog.Level
	msg   string
}

type loggerSpy struct {
	events []logHookEvent
}

func (ls *loggerSpy) Run(event *zerolog.Event, level zerolog.Level, msg string) {
	ls.events = append(ls.events, logHookEvent{event, level, msg})
}

func TestEsLoggerSupressesDeprecationWarnings(t *testing.T) {
	spy := loggerSpy{}
	testLogger := zerolog.New(os.Stdout)
	testLogger = testLogger.Hook(&spy)
	log := NewEsLogger(&testLogger, zerolog.InfoLevel)
	log.Printf("Deprecation warning: %s", fmt.Errorf("deprecation warning"))
	if len(spy.events) > 0 {
		t.FailNow()
	}
	log.Printf("not a deprecation warning %s", fmt.Errorf("some err"))
	if len(spy.events) == 0 {
		t.FailNow()
	}
}
