/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

// Package publisher handles publishing of LCT data to other receivers
//
//nolint:dupl // This needs more significant refactoring to eliminate the duplicate code
package publisher

import (
	"context"

	"github.com/rs/zerolog"
	elasticV6 "gopkg.in/olivere/elastic.v6"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
)

// ElasticSearchBulkProcessorsV6 creates a map of elastic search bulk processors for different clusters.
type ElasticSearchBulkProcessorsV6 struct {
	ClientsMap     *ElasticSearchClientsMap
	Log            *zerolog.Logger
	bulkProcessors map[string]*elasticV6.BulkProcessor
	Conf           *config.Configurations
}

func (e *ElasticSearchBulkProcessorsV6) getOrCreateProcessor(env string, afterFn elasticV6.BulkAfterFunc) *elasticV6.BulkProcessor {
	if e.bulkProcessors == nil {
		e.bulkProcessors = make(map[string]*elasticV6.BulkProcessor)
	}
	if e.bulkProcessors[env] == nil {
		e.Log.Warn().Msgf("Creating ES 6 bulk processor for env '%v'", env)
		proc, err := elasticV6.NewBulkProcessorService(e.ClientsMap.GetEsClientV6(env)).
			Workers(e.Conf.BulkProcessorWorkerThreads).
			BulkActions(e.Conf.EsBatchSizeEnvMap[env]).
			Stats(true).
			After(afterFn).
			Do(context.Background())
		if err != nil {
			e.Log.Fatal().Msgf("Failed to create bulk processor for env '%v'", env)
		} else {
			e.bulkProcessors[env] = proc
			err = proc.Start(context.Background())
			if err != nil {
				e.Log.Fatal().Msgf("Failed to start bulk processor for es env %v", env)
			}
		}
	}
	return e.bulkProcessors[env]
}

// Flush is exposed for testing purposes only.
func (e *ElasticSearchBulkProcessorsV6) Flush(cluster string) error {
	return e.bulkProcessors[cluster].Flush()
}
