/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package publisher

import (
	"fmt"
	"strings"

	"github.com/rs/zerolog"
	elasticV5 "gopkg.in/olivere/elastic.v5"
	elasticV6 "gopkg.in/olivere/elastic.v6"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/logger"
)

// ElasticSearchClientsMap -- maintains a map of Elasticsearch clients per cluster for different environments.
type ElasticSearchClientsMap struct {
	clientsV6 map[string]*elasticV6.Client
	clientsV5 map[string]*elasticV5.Client

	log *zerolog.Logger
}

// Initialize -- initializes clients per cluster for different environments.
func (e *ElasticSearchClientsMap) Initialize(conf *config.Configurations, log *zerolog.Logger, hostNamesMap map[string][]string) {
	e.clientsV5 = make(map[string]*elasticV5.Client)
	e.clientsV6 = make(map[string]*elasticV6.Client)
	e.log = log
	for env, hostnames := range hostNamesMap {
		esVersion := conf.EsVersionEnvMap[env]
		e.log.Debug().Msgf("ENV '%s', VERSION '%s'", env, esVersion)
		switch esVersion {
		case "es_6":
			e.log.Info().Msgf("Creating ES 6 client for env %v", env)
			e.log.Debug().Msgf("ES6 config: '%v'", conf)
			e.clientsV6[env] = e.createEsClientV6(env, hostnames, log, conf)
		default:
			e.log.Info().Msgf("Creating ES 5.6 client for env %v", env)
			e.clientsV5[env] = e.createEsClientV5(env, hostnames, log, conf)
		}
	}
}

// createEsClientV6 -- Create a v6 elasticsearch client
//
//nolint:dupl // This needs more significant refactoring to eliminate the duplicate code
func (e *ElasticSearchClientsMap) createEsClientV6(clusterName string, hostnames []string, log *zerolog.Logger, conf *config.Configurations) *elasticV6.Client {
	urls := make([]string, len(hostnames))
	for index, hostname := range hostnames {
		urls[index] = fmt.Sprintf("%v://%v", conf.EsScheme, hostname)
	}

	errorLogger := logger.NewEsLogger(log, zerolog.ErrorLevel)
	infoLogger := logger.NewEsLogger(log, zerolog.InfoLevel)
	traceLogger := logger.NewEsLogger(log, zerolog.DebugLevel)

	client, err := elasticV6.NewClient(elasticV6.SetSniff(conf.EsSniff), elasticV6.SetURL(urls...), elasticV6.SetInfoLog(infoLogger),
		elasticV6.SetErrorLog(errorLogger), elasticV6.SetTraceLog(traceLogger), elasticV6.SetScheme(conf.EsScheme),
		elasticV6.SetBasicAuth(conf.EsUser, conf.EsPassword))
	if err != nil {
		e.log.Fatal().Msgf("Failed to connect to cluster %v with urls %v, Error = %v", clusterName, strings.Join(urls, ","), err)
	}
	return client
}

// createEsClientV5 -- Create a v5 elasticsearch client
//
//nolint:dupl // This needs more significant refactoring to eliminate the duplicate code
func (e *ElasticSearchClientsMap) createEsClientV5(clusterName string, hostnames []string, log *zerolog.Logger, conf *config.Configurations) *elasticV5.Client {
	urls := make([]string, len(hostnames))
	for index, hostname := range hostnames {
		urls[index] = fmt.Sprintf("%v://%v", conf.EsScheme, hostname)
	}
	e.log.Info().Msgf("Using URLs: %v", urls)

	errorLogger := logger.NewEsLogger(log, zerolog.ErrorLevel)
	infoLogger := logger.NewEsLogger(log, zerolog.InfoLevel)
	traceLogger := logger.NewEsLogger(log, zerolog.DebugLevel)

	client, err := elasticV5.NewClient(elasticV5.SetSniff(conf.EsSniff), elasticV5.SetURL(urls...), elasticV5.SetInfoLog(infoLogger),
		elasticV5.SetErrorLog(errorLogger), elasticV5.SetTraceLog(traceLogger), elasticV5.SetScheme(conf.EsScheme))
	if err != nil {
		e.log.Fatal().Msgf("Failed to connect to cluster '%v' with urls %v, Error = %v", clusterName, strings.Join(urls, ","), err)
	}
	return client
}

// GetEsClientV5 selects esClient from map based on cluster
func (e *ElasticSearchClientsMap) GetEsClientV5(clusterName string) *elasticV5.Client {
	return e.clientsV5[clusterName]
}

// GetEsClientV6 selects esClient from map based on cluster
func (e *ElasticSearchClientsMap) GetEsClientV6(clusterName string) *elasticV6.Client {
	return e.clientsV6[clusterName]
}
