/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package publisher

import (
	"os"
	"testing"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"

	"carbonblack.com/protocols/psc/ems/assetupdater"
	"github.com/aws/aws-sdk-go/aws"
	"github.com/rs/zerolog"
	"github.com/sirupsen/logrus"
	"github.com/spf13/viper"
	"github.com/stretchr/testify/assert"

	"gitlab.bit9.local/psckit/pscapp"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"

	"google.golang.org/protobuf/types/known/timestamppb"
)

const assetUpdaterStreamName = "asset-updater-input-stream"

var (
	testLogger     = zerolog.New(os.Stdout)
	metricsFactory = stats.NewPrometheusMetrticsFactory("kinesispublisher", &testLogger)
	kplMetrics     = NewKplMetrics(metricsFactory)
)

func TestKinesisPublisher_ProtoMarshalUnMarshal(t *testing.T) {
	viper.Set("dev", "true")
	app, err := pscapp.NewApp("test")
	if err != nil {
		logrus.Fatalf("Failed to initialize psc app..%v", err)
	}

	eventTimestamp := timestamppb.Now()
	lastContactField := make(map[string]*assetupdater.AssetDataChangeEvent_Field)
	lastContactField["lastContact"] = &assetupdater.AssetDataChangeEvent_Field{
		ChangeType: assetupdater.AssetDataChangeEvent_SET_VALUE,
		Values:     []string{"1670835095157"},
	}

	lctEvent := &assetupdater.AssetDataChangeEvent{
		Id:                   "6f07d533-f645-4262-b5ce-4a23438387c1",
		EventCreateTimestamp: eventTimestamp,
		PublisherId:          "psc:platform:ems:app:last_contact_time",
		AssetUrn:             "psc:device:J7G6DTLN:12345678",
		Type:                 assetupdater.AssetDataChangeEvent_UPDATE,
		Fields:               lastContactField,
	}
	encodedBytes := ProtoMarshal(lctEvent, &app.Logger)
	decodedProto := ProtoUnmarshal(encodedBytes, &app.Logger)

	assert.Equal(t, "6f07d533-f645-4262-b5ce-4a23438387c1", decodedProto.Id, "id must match after UnMarshal of Proto")
	assert.Equal(t, eventTimestamp.Seconds, decodedProto.EventCreateTimestamp.Seconds, "TimeStamp Seconds must match after UnMarshal of Proto")
	assert.Equal(t, eventTimestamp.Nanos, decodedProto.EventCreateTimestamp.Nanos, "TimeStamp Nanos must match after UnMarshal of Proto")
	assert.Equal(t, "psc:platform:ems:app:last_contact_time", decodedProto.PublisherId, "PublisherId must match after UnMarshal of Proto")
	assert.Equal(t, "psc:device:J7G6DTLN:12345678", decodedProto.AssetUrn, "AssetUrn must match after UnMarshal of Proto")
	assert.Equal(t, "1670835095157", decodedProto.Fields["lastContact"].Values[0], "lastContact field value must match after UnMarshal of Proto")
}

func TestKinesisPublisher_ProtoUnMarshal_Err(t *testing.T) {
	decodedProto := ProtoUnmarshal(nil, &testLogger)
	if decodedProto == nil {
		t.Fail()
	}
}

func TestKinesisPublisher_ProtoMarshal_Err(t *testing.T) {
	bytes := ProtoMarshal(nil, &testLogger)
	if bytes != nil {
		t.Fail()
	}
}

func TestKinesisPublisher_StreamExists_NoRegion_Err(t *testing.T) {
	viper.Set("dev", "true")
	app, err := pscapp.NewApp("test")
	if err != nil {
		logrus.Fatalf("Failed to initialize psc app..%v", err)
	}

	cfg := &config.Configurations{
		EnvCategory:             "local",
		AssetRegistryStreamName: assetUpdaterStreamName,
		AggregationBatchSize:    1,
		ProducerBatchCount:      1,
	}

	kinesisPublisher := NewKinesisPublisherWithMetrics(&app.Logger, cfg, kplMetrics)

	streamExists := kinesisPublisher.StreamExists(aws.String(assetUpdaterStreamName))
	assert.Equal(t, false, streamExists)
}

func TestKinesisPublisher_StreamExists_No_Token_Err(t *testing.T) {
	viper.Set("dev", "true")
	app, err := pscapp.NewApp("test")
	if err != nil {
		logrus.Fatalf("Failed to initialize psc app..%v", err)
	}

	cfg := &config.Configurations{
		RegionName:              "us-east-1",
		EnvCategory:             "local",
		AssetRegistryStreamName: assetUpdaterStreamName,
		AggregationBatchSize:    1,
		ProducerBatchCount:      1,
	}

	kinesisPublisher := NewKinesisPublisherWithMetrics(&app.Logger, cfg, kplMetrics)

	streamExists := kinesisPublisher.StreamExists(aws.String(assetUpdaterStreamName))
	assert.Equal(t, false, streamExists)
}

func TestKinesisPublisher_PutRecords_No_Token_Err(_ *testing.T) {
	viper.Set("dev", "true")
	app, err := pscapp.NewApp("test")
	if err != nil {
		logrus.Fatalf("Failed to initialize psc app..%v", err)
	}

	cfg := &config.Configurations{
		RegionName:              "us-east-1",
		EnvCategory:             "local",
		AssetRegistryStreamName: assetUpdaterStreamName,
		AggregationBatchSize:    1,
		ProducerBatchCount:      1,
	}

	kinesisPublisher := NewKinesisPublisherWithMetrics(&app.Logger, cfg, kplMetrics)
	changeRecords := prepareDataChangeRecords(5)

	eventPerEnv := make(map[string]int64)
	eventPerEnv["local"] = 5
	kinesisPublisher.PutRecords(changeRecords, eventPerEnv)
}

func prepareDataChangeRecords(i int) []*assetupdater.AssetDataChangeEvent {
	var lctEvents []*assetupdater.AssetDataChangeEvent

	for i >= 0 {
		eventTimestamp := timestamppb.Now()
		lastContactField := make(map[string]*assetupdater.AssetDataChangeEvent_Field)
		lastContactField["lastContact"] = &assetupdater.AssetDataChangeEvent_Field{
			ChangeType: assetupdater.AssetDataChangeEvent_SET_VALUE,
			Values:     []string{"1670835095157"},
		}

		changeEvent := &assetupdater.AssetDataChangeEvent{
			Id:                   "6f07d533-f645-4262-b5ce-4a23438387c1",
			EventCreateTimestamp: eventTimestamp,
			PublisherId:          "psc:platform:ems:app:last_contact_time",
			AssetUrn:             "psc:device:J7G6DTLN:12345678",
			Type:                 assetupdater.AssetDataChangeEvent_UPDATE,
			Fields:               lastContactField,
		}
		lctEvents = append(lctEvents, changeEvent)
		i--
	}

	return lctEvents
}
