/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package publisher

import (
	"time"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"github.com/rs/zerolog"
	"gitlab.bit9.local/psc/platform/orgkey-go/v3"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"
)

// Publisher publishes last contact time messages throttling messages based on last published message time tracked by publisher.
// This is a generic Publisher and can be provided with a sink function to create Publisher for elastic search for now
// and some other data store later.
type Publisher struct {
	log                       *zerolog.Logger
	deviceLastPublishTime     map[string]map[DeviceKey]time.Time
	orgKeyCache               map[string]*orgkey.OrgKey
	deviceSuppressionTimeSecs map[string]time.Duration
	batchSize                 int
	channel                   chan *LctBatch
	currentBatch              *LctBatch
	lastCleanupRun            time.Time
	CleanUpRunInterval        time.Duration
	DeviceDataExpiryInterval  time.Duration
	suppressionCount          int64
	*PublisherMetrics
}

type PublisherMetrics struct {
	dedupCount     stats.MetricUpdateFunc
	devicesTracked stats.MetricUpdateFunc
}

func NewPublisherMetrics(factory *stats.PrometheusMetricsFactory) *PublisherMetrics {
	return &PublisherMetrics{
		dedupCount:     factory.NewCounterVec("dedup_count", "pub", []string{"device_env"}),
		devicesTracked: factory.NewGaugeVec("devices_tracked", "pub", []string{"device_env"}),
	}
}

// OnUpdate is called from kinesis consumer for this publisher to handle updates. This function processes the updates only
// if the lastContactTime in the cached entry is older than the suppression time or does not exist.
// This is to reject the messages that arrive within the threshold interval to ensure that we are not flooding the downstream system.
func (p *Publisher) OnUpdate(data *devicelastcontacttime.DeviceLastContactTime) {
	orgKey, err := p.getOrgKey(data)
	if err != nil {
		p.log.Error().Msgf("Failed to decode orgKey %v", data.OrgKey)
		return
	}
	deviceKey := DeviceKey{
		OrgID:    data.OrgKey,
		DeviceID: data.DeviceId,
	}

	env := orgKey.Env()
	if p.deviceLastPublishTime[env] == nil {
		// new environment seen, create device key mapping for it
		p.deviceLastPublishTime[env] = make(map[DeviceKey]time.Time)
	}

	lctTime := time.Unix(0, data.LastContactTimestamp*int64(time.Millisecond))

	// Only if the current time is outside the defined suppression for the environment do we update the
	// last contact time for this device.
	if lctTime.Sub(p.deviceLastPublishTime[env][deviceKey]) > p.deviceSuppressionTimeSecs[env] {
		p.currentBatch.DeviceDataList = append(p.currentBatch.DeviceDataList, data)
		// We are updating lct to deviceLastPublishTime map without confirming that it's flushed to the sink.
		// This is required so that further updates on same key are suppressed till the threshold lctTime interval.
		// It's also ok if the push to sink fails as we will get the message in next batch.
		p.deviceLastPublishTime[env][deviceKey] = lctTime
		p.devicesTracked(float64(len(p.deviceLastPublishTime[env])), env)
	} else { // Otherwise, within suppression time
		p.suppressionCount++
		p.log.Debug().Msgf("Device suppression count = %d", p.suppressionCount)
		p.dedupCount(1, env)
	}

	// If we hit the batch size, push what we have to the consumer channel and create a new batch
	if len(p.currentBatch.DeviceDataList) >= p.batchSize {
		// Add current batch to consumer Channel.
		p.channel <- p.currentBatch
		p.currentBatch = new(LctBatch)
	}
	p.cleanUp()
}

// getOrgKey -- Get the orgKey string from the data and return a decode orgkey object with env and orgId
func (p *Publisher) getOrgKey(data *devicelastcontacttime.DeviceLastContactTime) (*orgkey.OrgKey, error) {
	if p.orgKeyCache[data.OrgKey] == nil {
		val, err := orgkey.Decode(data.OrgKey)
		if err != nil {
			return nil, err
		}
		p.orgKeyCache[data.OrgKey] = val
	}
	orgKey := p.orgKeyCache[data.OrgKey]
	return orgKey, nil
}

// cleanUp -- Once every cleanup interval, iterates over the deviceLctMap and cleans up devices older than the defined purge time.
//   - CleanUpRunInterval: defined interval to run the cleanup
//   - DeviceDataExpiryInterval: age at which to remove the entry
func (p *Publisher) cleanUp() {
	// Refresh the map every 24 hrs (default) to avoid memory build up due to deleted/de-registered devices
	deletedCount := 0
	if time.Since(p.lastCleanupRun) > p.CleanUpRunInterval {
		t := time.Now()
		p.log.Warn().Msgf("Running cleanup to remove old devices from map...")
		for env, deviceMap := range p.deviceLastPublishTime {
			for deviceKey, value := range deviceMap {
				if time.Since(value) > p.DeviceDataExpiryInterval {
					p.log.Debug().Msgf("Removing %+v ...", deviceKey)
					delete(p.deviceLastPublishTime[env], deviceKey)
					deletedCount++
				}
			}
		}
		p.lastCleanupRun = time.Now()
		p.log.Warn().Msgf("Done running cleanup. Devices removed = %v, Time taken - %v ", deletedCount, time.Since(t))
	}
}

// Config -- config for publisher
type Config struct {
	Log             *zerolog.Logger
	BatchSize       int
	SuppressionTime map[string]time.Duration
	ChannelSize     int
	Sinkf           func(<-chan *LctBatch)
	TestMode        bool
}

// NewPublisher creates a publisher with provided batch size, thresholdTime duration for which messages gets rejected
// if received within the duration on same key and a sink function to consume the batch and push it to external data source.
func NewPublisher(config *Config, publisherMetrics *PublisherMetrics) *Publisher {
	p := &Publisher{
		log:                       config.Log,
		deviceLastPublishTime:     make(map[string]map[DeviceKey]time.Time),
		orgKeyCache:               make(map[string]*orgkey.OrgKey),
		deviceSuppressionTimeSecs: config.SuppressionTime,
		batchSize:                 config.BatchSize,
		channel:                   make(chan *LctBatch, config.ChannelSize),
		currentBatch:              new(LctBatch),
		lastCleanupRun:            time.Now(),
		CleanUpRunInterval:        time.Hour * 12,
		DeviceDataExpiryInterval:  time.Hour * 24,
		PublisherMetrics:          publisherMetrics,
	}
	go config.Sinkf(p.channel)
	return p
}

// Close closes the publisher Channel and flush ticker
func (p *Publisher) Close() {
	close(p.channel)
}

// GetLastPublishTime For testing purpose only
func (p *Publisher) GetLastPublishTime(deviceKey DeviceKey) time.Time {
	return p.deviceLastPublishTime["local"][deviceKey]
}

// LctBatch batch of messages to be pushed to external data source.
type LctBatch struct {
	DeviceDataList []*devicelastcontacttime.DeviceLastContactTime
}

// DeviceKey key for device data
type DeviceKey struct {
	OrgID    string
	DeviceID int64
}

// DeviceExists -- check for existence of a given device entry for a given environment
func (p *Publisher) DeviceExists(env string, key DeviceKey) bool {
	_, ok := p.deviceLastPublishTime[env]
	if ok {
		_, ok2 := p.deviceLastPublishTime[env][key]
		return ok2
	}
	return false
}
