/*
 * Copyright (c) 2023-2024 VMware Carbon Black, all rights reserved
 */

package publisher

import (
	"sync"
	"testing"
	"time"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"github.com/sirupsen/logrus"
	"github.com/spf13/viper"

	"gitlab.bit9.local/psc/platform/orgkey-go/v3"
	"gitlab.bit9.local/psckit/pscapp"
)

var (
	orgKey, _        = orgkey.Encode("local", 2023)
	publisherMetrics = NewPublisherMetrics(stats.NewPrometheusMetrticsFactory("publisher", &testLogger))
)

type DataSink struct {
	SinkData []*devicelastcontacttime.DeviceLastContactTime
	mu       *sync.Mutex
}

func defineApp(dataSink *DataSink, suppressMilli time.Duration, batchSize int) *Publisher {
	viper.Set("dev", "true")
	viper.Set("es-batch-size", "local@1000")
	viper.Set("device-suppression-time-secs", "local@100")
	viper.Set("es-versions", "local@es_6")
	app, err := pscapp.NewApp("test")
	if err != nil {
		logrus.Fatalf("Failed to initialize psc app..%v", err)
	}

	suppressionTimeMap := make(map[string]time.Duration)
	suppressionTimeMap["local"] = time.Millisecond * suppressMilli
	p := NewPublisher(&Config{
		Log:             &app.Logger,
		BatchSize:       batchSize,
		SuppressionTime: suppressionTimeMap,
		ChannelSize:     10,
		Sinkf:           dataSink.collectingSink,
	}, publisherMetrics)

	return p
}

// TestEsPublisher_OnUpdate -- Validate that last contact time updates as expected.
func TestEsPublisher_OnUpdate(t *testing.T) {
	dataSink := DataSink{mu: &sync.Mutex{}}
	pub := defineApp(&dataSink, 500, 1)

	// Device definition
	key := DeviceKey{OrgID: orgKey.String(), DeviceID: 123}

	logrus.Info("Check 1: last contact time updates when device is not defined yet")
	testTime := time.Now()
	data0 := devicelastcontacttime.DeviceLastContactTime{
		OrgKey:               key.OrgID,
		DeviceId:             key.DeviceID,
		LastContactTimestamp: testTime.UnixMilli(),
		DeregisterAfterTime:  0,
	}
	checkTime := time.Unix(0, data0.LastContactTimestamp*int64(time.Millisecond))

	pub.OnUpdate(&data0)
	if pub.GetLastPublishTime(key) != checkTime {
		t.Errorf(`FAILED: LastContactPublisher.GetLastPublishTime(key) %v != data0.LastContactTime %v`,
			pub.GetLastPublishTime(key), checkTime)
	}
	time.Sleep(time.Millisecond * 100) // Give it time to sink in and ensure we have 1 entry
	checkDataSinkSize(t, &dataSink, 1)

	logrus.Info("Check 2: last update time is within the suppression time of 500 millis so does not update")
	data1 := devicelastcontacttime.DeviceLastContactTime{
		OrgKey:               key.OrgID,
		DeviceId:             key.DeviceID,
		LastContactTimestamp: testTime.UnixMilli() + 100,
		DeregisterAfterTime:  0,
	}

	// update time is only 100 millis later, within the suppression time of 500 millis, so LCT should be unchanged
	pub.OnUpdate(&data1)
	if pub.GetLastPublishTime(key) != checkTime {
		t.Errorf(`FAILED: LastContactPublisher.GetLastPublishTime(key) %v != data0.LastContactTime %v`,
			pub.GetLastPublishTime(key), checkTime)
	}
	time.Sleep(time.Millisecond * 100) // Give it time to sink in and ensure we still only have 1 entry
	checkDataSinkSize(t, &dataSink, 1)

	logrus.Info("Check 3: last update time outside the suppression time frame")
	data2 := devicelastcontacttime.DeviceLastContactTime{
		OrgKey:               key.OrgID,
		DeviceId:             key.DeviceID,
		LastContactTimestamp: testTime.UnixMilli() + 600, // 600 millis past the threshold
		DeregisterAfterTime:  0,
	}
	pub.OnUpdate(&data2)
	checkTime = time.Unix(0, data2.LastContactTimestamp*int64(time.Millisecond))

	// update time is now outside the suppression time, so LCT should be updated
	if pub.GetLastPublishTime(key) != checkTime {
		t.Errorf(`FAILED: LastContactPublisher.GetLastPublishTime(key) %v != data2.LastContactTime %v`,
			pub.GetLastPublishTime(key), checkTime)
	}
	time.Sleep(time.Millisecond * 100) // Give it time to sink in and ensure we have 2 entries
	checkDataSinkSize(t, &dataSink, 2)
	pub.Close()
}

// TestEsPublisher_OnUpdate2 -- verifies that the batch handling with sink works as expected
func TestEsPublisher_OnUpdate2(t *testing.T) {
	dataSink := DataSink{mu: &sync.Mutex{}}
	pub := defineApp(&dataSink, 1, 5)

	now := time.Now().UnixMilli()

	for i := 0; i < 10; i++ {
		// Test1
		data0 := devicelastcontacttime.DeviceLastContactTime{
			OrgKey:               orgKey.String(),
			DeviceId:             int64(i),
			LastContactTimestamp: now,
			DeregisterAfterTime:  0,
		}
		pub.OnUpdate(&data0)
		switch i {
		case 0, 1, 2, 3:
			checkDataSinkSize(t, &dataSink, 0)
		case 4, 5, 6, 7, 8:
			if i == 4 {
				time.Sleep(1 * time.Millisecond)
			}
			checkDataSinkSize(t, &dataSink, 5)
		case 9:
			time.Sleep(1 * time.Millisecond)
			checkDataSinkSize(t, &dataSink, 10)
		}
	}
	pub.Close()
}

// TestEsPublisher_Cleanup -- Verify that publisher entries are properly removed from the memory structure once
// they have aged past our defined threshold time
func TestEsPublisher_Cleanup(t *testing.T) {
	dataSink := DataSink{mu: &sync.Mutex{}}
	pub := defineApp(&dataSink, 1, 1)
	pub.CleanUpRunInterval = time.Millisecond * 10 // run every 10 millis
	pub.DeviceDataExpiryInterval = time.Hour * 4   // remove entries older than 4 hours
	now := time.Now().UnixMilli()

	// Device Info
	key := DeviceKey{OrgID: orgKey.String(), DeviceID: 456}

	logrus.Info("Check 1:  define initial device (will be removed due to age")
	data0 := devicelastcontacttime.DeviceLastContactTime{
		OrgKey:               key.OrgID,
		DeviceId:             key.DeviceID,
		LastContactTimestamp: now - 5*60*60*1000, // 5 hours ago
		DeregisterAfterTime:  0,
	}
	checkTime := time.Unix(0, data0.LastContactTimestamp*int64(time.Millisecond))

	pub.OnUpdate(&data0)
	if pub.GetLastPublishTime(key) != checkTime {
		t.Errorf(`LastContactPublisher.GetLastPublishTime(key) %v != data0.LastContactTime %v`,
			pub.GetLastPublishTime(key).UnixMilli(), checkTime)
	}
	time.Sleep(time.Millisecond)
	checkDataSinkSize(t, &dataSink, 1)

	// Sleep, so the next on update call triggers cleanup
	time.Sleep(time.Millisecond * 20)

	// Test2  -- add another device to trigger cleanup
	logrus.Info("Check 2:  add another device to trigger cleanup")
	data1 := devicelastcontacttime.DeviceLastContactTime{
		OrgKey:               key.OrgID,
		DeviceId:             4560,
		LastContactTimestamp: now + 100,
		DeregisterAfterTime:  0,
	}
	pub.OnUpdate(&data1)

	// check publisher structure to ensure that device 1 was removed
	if pub.DeviceExists("local", key) {
		t.Error(`Cleanup failed..`)
	}
	time.Sleep(time.Millisecond)
	checkDataSinkSize(t, &dataSink, 2)
	pub.Close()
}

// func TestEsPublisher_OrgMetrics(t *testing.T) {
//	viper.Set("dev", "true")
//	app, err := pscapp.NewApp("test")
//	if err != nil {
//		logrus.Fatalf("Failed to initialize psc app.%v", err)
//	}
//	dataSink := DataSink{mu: &sync.Mutex{}}
//	suppressionTimeMap := make(map[string]time.Duration)
//	suppressionTimeMap["local"] = time.Millisecond
//	p := publisher.NewPublisher(&publisher.Config{
//		Log:             &app.Logger,
//		BatchSize:       1,
//		SuppressionTime: suppressionTimeMap,
//		ChannelSize:     10,
//		Sinkf:           dataSink.collectingSink}, app.MetricsRegistry)
//	p.CleanUpRunInterval = time.Millisecond * 10
//	p.DeviceDataExpiryInterval = time.Minute * 10
//	now := time.Now().Unix() * 1000
//
//	// Test1
//	orgKey, err := orgkey.Encode("local", 1)
//	if err != nil {
//		t.Error(`Failed to orgKey`)
//	}
//	data0 := devicelastcontacttime.DeviceLastContactTime{
//		OrgID:               orgKey.String(),
//		DeviceId:             1,
//		LastContactTimestamp: now - 5*60*1000,
//		DeregisterAfterTime:  0,
//	}
//
//	key := publisher.DeviceKey{
//		OrgID:    data0.OrgID,
//		DeviceID: data0.DeviceId,
//	}
//	p.OnUpdate(&data0)
//	if p.GetLastPublishTime(key) != time.Unix(0, data0.LastContactTimestamp*int64(time.Millisecond)) {
//		t.Error(`LastContactPublisher.GetLastPublishTime(key) != data0.LastContactTime`)
//	}
//	time.Sleep(time.Millisecond)
//	checkDataSinkSize(t, &dataSink, 1)
//
//	// Sleep for 10 millis, so the next on update call triggers cleanup
//	time.Sleep(time.Millisecond * 10)
//
//	// Test2
//	data1 := devicelastcontacttime.DeviceLastContactTime{
//		OrgID:               orgKey.String(),
//		DeviceId:             2,
//		LastContactTimestamp: now + 100,
//		DeregisterAfterTime:  0,
//	}
//	p.OnUpdate(&data1)
//	time.Sleep(time.Millisecond)
//	checkDataSinkSize(t, &dataSink, 2)
//	tags := make(map[string]string)
//	tags["env"] = "local"
//	tags["org"] = orgKey.String()
//	gauge, ok := app.MetricsRegistry.GetMetric("psc.last_contact_updater.lct_time_10m", tags).(metrics.Gauge)
//	if !ok {
//		t.Error("Error while fetching metrics")
//	}
//	if gauge.Value() != 2 {
//		t.Error("Metrics value not correct.")
//	}
//	p.Close()
//}

func checkDataSinkSize(t *testing.T, dataSink *DataSink, size int) {
	dataSink.mu.Lock()
	if len(dataSink.SinkData) != size {
		t.Errorf("FAILED: len(dataSink.SinkData) %d != %d", len(dataSink.SinkData), size)
	}
	dataSink.mu.Unlock()
}

func (d *DataSink) collectingSink(sinkChannel <-chan *LctBatch) {
	for b := range sinkChannel {
		d.mu.Lock()
		d.SinkData = append(d.SinkData, b.DeviceDataList...)
		d.mu.Unlock()
	}
}
