package stats

import (
	"fmt"
	"net/http"
	"os"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/collectors"
	"github.com/prometheus/client_golang/prometheus/promauto"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"github.com/rs/zerolog"
)

const (
	metricsPrefix = "ems_last_contact_updater_"
	envLabel      = "env"
	sourceLabel   = "source"
)

type PrometheusMetricsFactory struct {
	gatherer   prometheus.Gatherer
	registerer prometheus.Registerer
	factory    *promauto.Factory
	log        *zerolog.Logger
}

func NewPrometheusMetrticsFactory(envName string, log *zerolog.Logger) *PrometheusMetricsFactory {
	return newPrefixWrappedRegistry(metricsPrefix, envName, log)
}

func newPrefixWrappedRegistry(prefix, env string, log *zerolog.Logger) *PrometheusMetricsFactory {
	registry := prometheus.NewRegistry()
	wrapped := prometheus.WrapRegistererWithPrefix(prefix, registry)
	wrapped = prometheus.WrapRegistererWith(prometheus.Labels{
		envLabel: env, sourceLabel: func() string {
			hostname, err := os.Hostname()
			if err != nil {
				hostname = "unknown"
			}

			return hostname
		}(),
	}, wrapped)
	promFactory := promauto.With(wrapped)
	return &PrometheusMetricsFactory{registerer: wrapped, gatherer: registry, factory: &promFactory, log: log}
}

func getMetricsOpts(name, subsystem string) prometheus.Opts {
	return prometheus.Opts{Name: name, Help: name, Subsystem: subsystem}
}

func (f *PrometheusMetricsFactory) NewCounterVec(name, subsystem string, tags []string) MetricUpdateFunc {
	counterVec := f.factory.NewCounterVec(prometheus.CounterOpts(getMetricsOpts(name, subsystem)), tags)
	return func(v float64, tagValues ...string) {
		counter, err := counterVec.GetMetricWithLabelValues(tagValues...)
		if err != nil {
			f.log.Err(fmt.Errorf("failure to update metric %s with tag-values %v: %w", name, tagValues, err))
			return
		}
		counter.Add(v)
	}
}

func (f *PrometheusMetricsFactory) NewGaugeVec(name, subsystem string, tags []string) MetricUpdateFunc {
	gaugeVec := f.factory.NewGaugeVec(prometheus.GaugeOpts(getMetricsOpts(name, subsystem)), tags)
	return func(v float64, tagValues ...string) {
		gauge, err := gaugeVec.GetMetricWithLabelValues(tagValues...)
		if err != nil {
			f.log.Err(fmt.Errorf("failure to update metric %s with tag-values %v: %w", name, tagValues, err))
			return
		}
		gauge.Set(v)
	}
}

func (f *PrometheusMetricsFactory) HttpHandler() http.Handler {
	return promhttp.HandlerFor(f.gatherer, promhttp.HandlerOpts{})
}

func (f *PrometheusMetricsFactory) RegisterGoMetrics() {
	f.registerer.MustRegister(collectors.NewGoCollector(
		collectors.WithGoCollectorRuntimeMetrics(collectors.MetricsMemory),
	))
}

type MetricUpdateFunc func(value float64, tagValues ...string)
